package TDS_Calculate;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.text.DecimalFormat;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.Period;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;

/**
 *
 * @author 1121947
 */
public class TDS_Calculation {

    public static final String URL = "jdbc:oracle:thin:@localhost:1521:IPKSDB";
    public static final String USERNAME = "pacs_db";
    public static final String PASWORD = "pacs_db";
    private static final int batch_length = 1000;

    public static void main(String[] args) {
        //get the TD/RD account level information
        List<Map<String, Object>> depAccountsDetails = getDepAccountsDetails();

        int hSize = depAccountsDetails.size();
        //System.out.println(hSize);

        //Get the PAN Data
        Map<String, ID_Type> identificationData = fetchPANDetails(depAccountsDetails);
        //print the fetched data
//        for (Map.Entry<String, String> panDetails : identificationData.entrySet()) {
//            System.out.println("cif_no: " + panDetails.getKey() + " having ID Type:" + panDetails.getValue());
//        }

        //Get the 15GH/PAN details as per the customerNumber from the Previous method
        checkIdentificationAvailability(depAccountsDetails, identificationData);

        //Calculate the Projected value and insert into the table TDS_TXN
       // for (Map<String, Object> accountDetails : depAccountsDetails) {
            //String accNo = (String) accountDetails.get("KEY_1");

            //Caclculate the Projected amount.
           /// calculateProjectedAmount(accountDetails);
            //System.out.println("Projected amount for the account:" + accNo + "  " + ProjectedAmt);
            //insertProjectedAmount(accNo, ProjectedAmt, prodCode);
      //  }

    }

    public static List<Map<String, Object>> getDepAccountsDetails() {
        List<Map<String, Object>> depAccountsDetails = new ArrayList<>();

        try (Connection conn = DriverManager.getConnection(URL, USERNAME, PASWORD)) {

            String depAccountSql = "SELECT DA.KEY_1,DA.PACS_ID,DA.CURR_STATUS,DA.CUSTOMER_NO,DA.ACCT_OPEN_DT,\n"
                    + "DA.INTT_PAID,DP.PROD_CODE,DA.INTT_SOP_DATE,DA.INTT_EOP_DATE,DA.TERM_VALUE,DA.MAT_VALUE,DA.MAT_DT,DA.TERM_LENGTH,\n"
                    + "DA.NXT_DUE_DATE,DA.INTT_PROJECTED,dp.INTT_PAYOUT_FREQ FROM DEP_ACCOUNT DA ,dep_product dp\n"
                    + "where DA.DEP_PROD_ID=dp.ID and substr(dp.prod_code,1,1) in ('2','3') and substr(DA.GL_CLASS_CODE,9,2)='14'\n"
                    + "and DA.CURR_STATUS ='O' and DA.PACS_ID in  ('07544') group by DA.KEY_1,DA.PACS_ID,DA.CURR_STATUS,DA.CUSTOMER_NO,DA.ACCT_OPEN_DT,\n"
                    + "DA.INTT_PAID,DA.INTT_SOP_DATE,DA.INTT_EOP_DATE,DA.TERM_VALUE,DA.MAT_VALUE,DA.MAT_DT,DA.TERM_LENGTH,\n"
                    + "DA.NXT_DUE_DATE,DA.INTT_PROJECTED,dp.INTT_PAYOUT_FREQ,DP.PROD_CODE";
            PreparedStatement preparedStatement = conn.prepareStatement(depAccountSql);
            ResultSet resultSet = preparedStatement.executeQuery();

            while (resultSet.next()) {
                Map<String, Object> accountDetail = new HashMap<>();
                accountDetail.put("KEY_1", resultSet.getString("KEY_1"));
                accountDetail.put("PACS_ID", resultSet.getString("PACS_ID"));
                accountDetail.put("CURR_STATUS", resultSet.getString("CURR_STATUS"));
                accountDetail.put("CUSTOMER_NO", resultSet.getString("CUSTOMER_NO"));
                accountDetail.put("ACCT_OPEN_DT", resultSet.getDate("ACCT_OPEN_DT"));
                accountDetail.put("INTT_PAID", resultSet.getDouble("INTT_PAID"));
                accountDetail.put("PROD_CODE", resultSet.getString("PROD_CODE"));
                accountDetail.put("INTT_SOP_DATE", resultSet.getDate("INTT_SOP_DATE"));
                accountDetail.put("INTT_EOP_DATE", resultSet.getDate("INTT_EOP_DATE"));
                accountDetail.put("TERM_VALUE", resultSet.getInt("TERM_VALUE"));
                accountDetail.put("MAT_VALUE", resultSet.getString("MAT_VALUE"));
                accountDetail.put("MAT_DT", resultSet.getDate("MAT_DT"));
                accountDetail.put("TERM_LENGTH", resultSet.getInt("TERM_LENGTH"));
                accountDetail.put("NXT_DUE_DATE", resultSet.getDate("NXT_DUE_DATE"));
                accountDetail.put("INTT_PROJECTED", resultSet.getDouble("INTT_PROJECTED"));
                accountDetail.put("INTT_PAYOUT_FREQ", resultSet.getString("INTT_PAYOUT_FREQ"));

                depAccountsDetails.add(accountDetail);

            }

        } catch (Exception Ex) {
            Ex.printStackTrace();
        }

        return depAccountsDetails;
    }

    private static void checkIdentificationAvailability(List<Map<String, Object>> depAccountsDetails, Map<String, ID_Type> identificationData) {

        String formType = null;
        String birth_date = null;

        try (Connection conn = DriverManager.getConnection(URL, USERNAME, PASWORD)) {
            for (Map<String, Object> acc : depAccountsDetails) {
                String CIF_No = (String) acc.get("CUSTOMER_NO");
                String KEY_1 = (String) acc.get("KEY_1");
                Double INTT_PAID = (Double) acc.get("INTT_PAID");
                String PACS_ID = (String) acc.get("PACS_ID");

                ID_Type id_data = identificationData.get(CIF_No);

                String id_num = id_data.getIdNo();
                String ID_TYPE = id_data.getIDType();

                System.out.println("ID TYPE is: " + ID_TYPE + " for cif no:" + CIF_No + " and ID_NO: " + id_num);

                //  LocalDate birthDt = null;
                String bDaySql = "SELECT TO_CHAR(TO_DATE(BIRTH_DATE,'DD-MM-YYYY'),'DD-MM-YYYY') AS BIRTH_DATE FROM KYC_HDR WHERE CIF_NO = ?";
                try (PreparedStatement ps = conn.prepareStatement(bDaySql)) {
                    ps.setString(1, CIF_No);
                    try (ResultSet rs = ps.executeQuery()) {

                        while (rs.next()) {

                            try {
                                birth_date = rs.getString("BIRTH_DATE");
                                //System.out.println("Birth Date: "+birth_date);
                            } catch (SQLException e) {
                                System.err.println("null pointer");
                            }
                            if (birth_date != null) {
                                //System.out.println("Birth Date:" + birth_date);
                                LocalDate submission_date = LocalDate.now();

                                DateTimeFormatter bdayFormatter = DateTimeFormatter.ofPattern("dd-MM-yyyy");
                                LocalDate birthdate = LocalDate.parse(birth_date, bdayFormatter);
                                //if (birth_date != null) {
                                Period period = Period.between(birthdate, submission_date);
                                int age = period.getYears(); // Get the years from here.

                                if (age >= 60) {
                                    formType = "H";
                                } else {
                                    formType = "G";
                                }

                                if ("PAN CARD".equalsIgnoreCase(ID_TYPE)) {
                                    PanAvailability(KEY_1, INTT_PAID, CIF_No, formType, PACS_ID);
                                } else {
                                    PanNotAvailability(KEY_1, INTT_PAID, CIF_No, formType, PACS_ID);
                                }
                            }
                        }
                    } catch (Exception e) {
                        //System.out.println("Processing done");
                        e.printStackTrace();
                    }
                } catch (Exception e) {
                    //System.out.println("Processing done");
                    e.printStackTrace();
                }

            }

        } catch (Exception ex) {
            ex.printStackTrace();
        }
    }

    public static void PanAvailability(String KEY_1, Double INTT_PAID, String CIF_No, String formType, String PACS_ID) {

        System.out.println("CIF No:" + CIF_No + " is having Pan " + " with Form Type " + formType + " and interest Paid:" + INTT_PAID + " for PACS ID: " + PACS_ID + "  of the a/c no: " + KEY_1);
        double thresold_G = 250000;
        double thresold_H = 300000;
        double thresold = 0.0;
        double tds_amount = 0.0;

        if (formType.equalsIgnoreCase("H")) {
            thresold = thresold_H;
        } else if (formType.equalsIgnoreCase("G")) {
            thresold = thresold_G;
        }

        DecimalFormat df_obj = new DecimalFormat("#.####");

        try (Connection conn = DriverManager.getConnection(URL, USERNAME, PASWORD)) {
//            for (Map<String, Object> acc : depAccountsDetails) {
//                String AccCIF_No = (String) acc.get("CUSTOMER_NO");
//
//                System.out.println("AccCIF No: " + AccCIF_No);
//               if (CIF_No.equals(AccCIF_No)) {
//                    String KEY_1 = (String) acc.get("KEY_1");
//                    Double INTT_PAID = (Double) acc.get("INTT_PAID");
//                   // double INTT_PROJECTED = (double) acc.get("INTT_PROJECTED");
//                    
//                   
//                    // Added for GL_TXN calculations
//                    String PACS_ID = (String) acc.get("PACS_ID");

            // extract the Form _id from FORM15G_H table.
            String getFormData = "SELECT FORM_ID,KEY_1 FROM FORM15G_H WHERE KEY_1= ?";
            PreparedStatement ps = conn.prepareStatement(getFormData);

            ps.setString(1, CIF_No);

            ResultSet formResultSet = ps.executeQuery();

            int FORM_ID = 0;

            if (formResultSet.next()) {
                FORM_ID = formResultSet.getInt("FORM_ID");
                System.out.println("Form Id is for YES PAN.........................." + FORM_ID);
            }

            // Extract the TXN_ID from the table TDX_TXN with the help of Key_1;
            String getTdsTxnData = "select TXN_ID from TDS_TXN where KEY_1= ?";
            PreparedStatement ps1 = conn.prepareStatement(getTdsTxnData);

            ps1.setString(1, KEY_1);

            ResultSet rs = ps1.executeQuery();

            int TXN_ID = 0;

            if (rs.next()) {
                TXN_ID = rs.getInt("TXN_ID");
            }

            //Including 15g/h logic.
            if (FORM_ID != 0) {
                System.out.println("FORM ID for having 15g/h for YES PAN:" + FORM_ID);
                // Starting of TDS Calculation........

                if (INTT_PAID != null && INTT_PAID > thresold) {

                    System.out.println("The account number: " + KEY_1 + " has been crossed the thresold:" + INTT_PAID + " having CIF No: " + CIF_No + " for YES PAN");

                    //Calculate the TDS.//Here TDS will be 10% as PAN is avaiable            
                    System.out.println("Interest Paid: " + INTT_PAID);
                    tds_amount = INTT_PAID * 0.10;
                    System.out.println("TDS Amount is : " + tds_amount);

                    // deduct the TDS_AMOUNT from INTT_PAID
                    double deductedAmt = INTT_PAID - tds_amount;

                    System.out.println("Deducted amount: " + deductedAmt);

                    // Now need to insert into the table for record.
                    String InsertTdsSql = " Insert into TDS_CALCULATIONS(TXN_ID,FORM_ID,TDS_AMT,CREATED_AT,TDS_DATE,KEY_1,INTEREST_AMT,PANCARD) values(?,?,?,?,?,?,?,?) ";
                    PreparedStatement insertSmt = conn.prepareStatement(InsertTdsSql);

                    insertSmt.setInt(1, TXN_ID);
                    insertSmt.setInt(2, FORM_ID);
                    insertSmt.setDouble(3, tds_amount);
                    insertSmt.setTimestamp(4, Timestamp.valueOf(LocalDateTime.now()));
                    insertSmt.setDate(5, java.sql.Date.valueOf(LocalDate.now()));
                    insertSmt.setString(6, KEY_1);
                    insertSmt.setDouble(7, INTT_PAID);
                    insertSmt.setString(8, "Y");

                    insertSmt.executeQuery();

                    // Need to insert the deducted amount into corresponding GL.
                     insrtGLTransaction(KEY_1, PACS_ID, INTT_PAID, tds_amount);
                } else {
                    System.out.println("Interest Paid didn't cross the thresold for YES PAN having FORM_ID:" + FORM_ID);
                }
            } else {
                //This block is for calculating TDS for more than 10000 due to 15g/h availability
                System.out.println("FORM ID for not having 15g/h:" + FORM_ID);

                if (formType.equalsIgnoreCase("H")) {
                    thresold = 50000;
                } else if (formType.equalsIgnoreCase("G")) {
                    thresold = 40000;
                }

                System.out.println("NO 15G/H available for YES PAN setting thresold: " + thresold);

                if (INTT_PAID != null && INTT_PAID > thresold) {
                    System.out.println("The account number: " + KEY_1 + " has crossed the thresold:" + INTT_PAID + " having CIF No: " + CIF_No);

                    //Calculate the TDS.//Here TDS will be 10% as PAN is avaiable            
                    System.out.println("Interest Paid: " + INTT_PAID);
                    tds_amount = INTT_PAID * 0.10;
                    System.out.println("TDS Amount is : " + tds_amount);

                    // deduct the TDS_AMOUNT from INTT_PAID
                    double deductedAmt = INTT_PAID - tds_amount;

                    System.out.println("Deducted amount: " + deductedAmt);

                    // Now need to insert into the table for record.
                    String InsertTdsSql = " Insert into TDS_CALCULATIONS(TXN_ID,FORM_ID,TDS_AMT,CREATED_AT,TDS_DATE,KEY_1,INTEREST_AMT,PANCARD) values(?,?,?,?,?,?,?,?) ";
                    PreparedStatement insertSmt = conn.prepareStatement(InsertTdsSql);

                    insertSmt.setInt(1, TXN_ID);
                    insertSmt.setInt(2, FORM_ID);
                    insertSmt.setDouble(3, tds_amount);
                    insertSmt.setTimestamp(4, Timestamp.valueOf(LocalDateTime.now()));
                    insertSmt.setDate(5, java.sql.Date.valueOf(LocalDate.now()));
                    insertSmt.setString(6, KEY_1);
                    insertSmt.setDouble(7, INTT_PAID);
                    insertSmt.setString(8, "Y");

                    insertSmt.executeQuery();

                    // Need to insert the deducted amount into corresponding GL.
                      insrtGLTransaction(KEY_1, PACS_ID, INTT_PAID, tds_amount);
                } else {
                    System.out.println("Interest Paid did not cross the thresold for YES PAN having FORM_ID:" + FORM_ID);
                }
            }
            //}

        } catch (Exception e) {
            e.printStackTrace();
        } finally {
            System.out.println("Processing done for PAN available");
            System.out.println("\n");
        }
    }

    public static void PanNotAvailability(String KEY_1, Double INTT_PAID, String CIF_No, String formType, String PACS_ID) {
        System.out.println("CIF No:" + CIF_No + " is not having Pan " + " with Form Type " + formType + " and interest Paid:" + INTT_PAID + "  for PACS ID: " + PACS_ID + "  of the a/c no: " + KEY_1);

        double thresold_G = 250000;
        double thresold_H = 300000;
        double thresold = 0;
        double tds_amount = 0;

        if (formType.equalsIgnoreCase("H")) {
            thresold = thresold_H;
        } else if (formType.equalsIgnoreCase("G")) {
            thresold = thresold_G;
        }

        try (Connection conn = DriverManager.getConnection(URL, USERNAME, PASWORD)) {
//            for (Map<String, Object> acc : depAccountsDetails) {
//                String AccCIF_No = (String) acc.get("CUSTOMER_NO");
//
//                if (CIF_No.equals(AccCIF_No)) {
//                    String KEY_1 = (String) acc.get("KEY_1");
//                    Double INTT_PAID = (Double) acc.get("INTT_PAID");
//                    double INTT_PROJECTED = (double) acc.get("INTT_PROJECTED");
//                   // thresold = formType.equals("H") ? thresold_H : thresold_G;
//                   
//                    
//                    // Added for GL_TXN calculations
//                    String PACS_ID = (String) acc.get("PACS_ID");

            // extract the Form _id from FORM15G_H table.
            String getFormData = "SELECT FORM_ID,KEY_1 FROM FORM15G_H WHERE KEY_1= ?";
            PreparedStatement ps = conn.prepareStatement(getFormData);

            ps.setString(1, CIF_No);

            ResultSet formResultSet = ps.executeQuery();

            int FORM_ID = 0;

            if (formResultSet.next()) {
                FORM_ID = formResultSet.getInt("FORM_ID");
                System.out.println("Form Id is for NO PAN..........................." + FORM_ID);
            }
            //System.out.println("FORM ID:" + FORM_ID);

            // Extract the TXN_ID from the table TDX_TXN with the help of Key_1;
            String getTdsTxnData = "select TXN_ID from TDS_TXN where KEY_1= ?";
            PreparedStatement ps1 = conn.prepareStatement(getTdsTxnData);

            ps1.setString(1, KEY_1);

            ResultSet rs = ps1.executeQuery();

            int TXN_ID = 0;

            if (rs.next()) {
                TXN_ID = rs.getInt("TXN_ID");
            }

            //Including 15g/h logic.
            if (FORM_ID != 0) {
                System.out.println("FORM ID for having 15g/h for NO PAN: " + FORM_ID);
                // Starting of TDS Calculation........

                if (INTT_PAID != null && INTT_PAID > thresold) {

                    System.out.println("The account number: " + KEY_1 + " has been crossed the thresold:" + INTT_PAID + " having CIF No: " + CIF_No + " for NO PAN");

                    //Calculate the TDS.//Here TDS will be 10% as PAN is avaiable            
                    System.out.println("Interest Paid: " + INTT_PAID);
                    tds_amount = INTT_PAID * 0.20;
                    System.out.println("TDS Amount is : " + tds_amount);

                    // deduct the TDS_AMOUNT from INTT_PAID
                    double deductedAmt = INTT_PAID - tds_amount;

                    System.out.println("Deducted amount: " + deductedAmt);

                    // Now need to insert into the table for record.
                    String InsertTdsSql = " Insert into TDS_CALCULATIONS(TXN_ID,FORM_ID,TDS_AMT,CREATED_AT,TDS_DATE,KEY_1,INTEREST_AMT,PANCARD) values(?,?,?,?,?,?,?,?) ";
                    PreparedStatement insertSmt = conn.prepareStatement(InsertTdsSql);

                    insertSmt.setInt(1, TXN_ID);
                    insertSmt.setInt(2, FORM_ID);
                    insertSmt.setDouble(3, tds_amount);
                    insertSmt.setTimestamp(4, Timestamp.valueOf(LocalDateTime.now()));
                    insertSmt.setDate(5, java.sql.Date.valueOf(LocalDate.now()));
                    insertSmt.setString(6, KEY_1);
                    insertSmt.setDouble(7, INTT_PAID);
                    insertSmt.setString(8, "Y");

                    insertSmt.executeQuery();

                    // Need to insert the deducted amount into corresponding GL.
                     insrtGLTransaction(KEY_1, PACS_ID, INTT_PAID, tds_amount);
                } else {
                    System.out.println("Interest Paid did not cross the thresold for NO PAN having FORM_ID:" + FORM_ID);
                }
            } else {
                //This block is for calculating TDS for more than 10000 due to 15g/h availability
                System.out.println("FORM ID for not having 15g/h:" + FORM_ID);
                System.out.println("NO 15G/H available for NO PAN");

                if (formType.equalsIgnoreCase("H")) {
                    thresold = 50000;
                } else if (formType.equalsIgnoreCase("G")) {
                    thresold = 40000;
                }

                if (INTT_PAID != null && INTT_PAID > thresold) {
                    System.out.println("The account number: " + KEY_1 + " has crossed the thresold:" + INTT_PAID + " having CIF No: " + CIF_No + " for NO PAN");

                    //Calculate the TDS.//Here TDS will be 10% as PAN is avaiable            
                    System.out.println("Interest Paid: " + INTT_PAID);
                    tds_amount = INTT_PAID * 0.20;
                    System.out.println("TDS Amount is : " + tds_amount);

                    // deduct the TDS_AMOUNT from INTT_PAID
                    double deductedAmt = INTT_PAID - tds_amount;

                    System.out.println("Deducted amount: " + deductedAmt);

                    // Now need to insert into the table for record.
                    String InsertTdsSql = " Insert into TDS_CALCULATIONS(TXN_ID,FORM_ID,TDS_AMT,CREATED_AT,TDS_DATE,KEY_1,INTEREST_AMT,PANCARD) values(?,?,?,?,?,?,?,?) ";
                    PreparedStatement insertSmt = conn.prepareStatement(InsertTdsSql);

                    insertSmt.setInt(1, TXN_ID);
                    insertSmt.setInt(2, FORM_ID);
                    insertSmt.setDouble(3, tds_amount);
                    insertSmt.setTimestamp(4, Timestamp.valueOf(LocalDateTime.now()));
                    insertSmt.setDate(5, java.sql.Date.valueOf(LocalDate.now()));
                    insertSmt.setString(6, KEY_1);
                    insertSmt.setDouble(7, INTT_PAID);
                    insertSmt.setString(8, "Y");

                    insertSmt.executeQuery();

                    // Need to insert the deducted amount into corresponding GL.
                     insrtGLTransaction(KEY_1, PACS_ID, INTT_PAID, tds_amount);
                } else {
                    System.out.println("Interest Paid did not cross the thresold for NO PAN having FORM_ID:" + FORM_ID);
                }
            }
            //}

        } catch (Exception e) {
            e.printStackTrace();
        } finally {
            System.out.println("Processing done for PAN NOT available");
            System.out.println("\n");
        }
    }

    public static void calculateProjectedAmount(Map<String, Object> accNo) {

        //  int termValue = (int) accNo.get("TERM_VALUE");
        double INTT_PROJECTED = (double) accNo.get("INTT_PROJECTED");
        // int termLength = (int) accNo.get("TERM_LENGTH"); //in days as per IPKS DB
        String inttPayoutFre = ((String) accNo.get("INTT_PAYOUT_FREQ")).trim();
        String prodCode = (String) accNo.get("PROD_CODE");
        String accNum = (String) accNo.get("KEY_1");
        int temp = 0;

        switch (inttPayoutFre) {
            case "M":
                temp = 12;
                break;
            case "Q":
                temp = 4;
                break;
            case "Y":
                temp = 1;
                break;
            case "O":
                temp = 2; // need to be changed here.
                break;
            default:
                throw new IllegalArgumentException("Invalid interest payout frequency: " + inttPayoutFre);

        }

        //calculate projected amount
        double projectedAmount = INTT_PROJECTED * temp;
        System.out.println("Calculation is done for Account No: " + accNum + " and Projected Amount is : " + projectedAmount);
        // return projectedAmount;
        insertProjectedAmount(accNum, projectedAmount, prodCode);

    }

    public static void insertProjectedAmount(String accNo, double projectedAmt, String prodCode) {
        String temp = null;

        try (Connection conn = DriverManager.getConnection(URL, USERNAME, PASWORD)) {

            if (prodCode.charAt(0) == '2') {
                temp = "TD";
            } else if (prodCode.charAt(0) == '3') {
                temp = "RD";
            }

            String insertTxnSql = "insert into TDS_TXN (KEY_1,TXN_DATE,TXN_TYPE,CREATED_AT,PROJ_AMT) values(?,?,?,?,?)";

            PreparedStatement ps = conn.prepareStatement(insertTxnSql);

            ps.setString(1, accNo);
            ps.setDate(2, java.sql.Date.valueOf(LocalDate.now()));
            ps.setString(3, temp);
            ps.setTimestamp(4, Timestamp.valueOf(LocalDateTime.now()));
            ps.setDouble(5, projectedAmt);

            boolean i = ps.executeUpdate() > 0;
            System.out.println("inserted Projected amount for : " + accNo);

        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    private static Map<String, ID_Type> fetchPANDetails(List<Map<String, Object>> depAccountsDetails) {

        Map<String, ID_Type> PanData = new HashMap<>();

        List<String> cif_No = new ArrayList<>();
        //  String formType = null;

        for (Map<String, Object> acc : depAccountsDetails) {
            cif_No.add((String) acc.get("CUSTOMER_NO"));
        }

        int totalrecords = cif_No.size();
        System.out.println("Total:" + totalrecords);
        //Boolean panavailable = false;
        int numOfBatches = (totalrecords / batch_length) + (totalrecords % batch_length == 0 ? 0 : 1);
        System.out.println(numOfBatches);

        try (Connection conn = DriverManager.getConnection(URL, USERNAME, PASWORD)) {

            for (int batch = 0; batch < numOfBatches; batch++) {
                int startIndex = batch * batch_length;
                int endIndex = Math.min(startIndex + batch_length, totalrecords);
                // System.out.println("Start:" + startIndex + "end:" + endIndex);

                List<String> batchCifNos = cif_No.subList(startIndex, endIndex);
                //System.out.println("batch cif no:" + batchCifNos);

                //Create a list of CIF No.
                StringBuilder stringbuilder = new StringBuilder("SELECT KH.CIF_NO as CIF_NO,KD.ID_NO as ID_NO,m.DOCUMENT_DTL as ID_TYPE,KH.BIRTH_DATE as BIRTH_DATE from KYC_HDR KH,"
                        + "PACS_MASTER PM,kyc_dtl KD,kyc_document_mst m where KH.PACS_ID=PM.PACS_ID and KH.CIF_NO = KD.CIF_NO and KD.kyc_document_mst_id = m.id AND KH.CIF_NO in (");
                for (int i = 0; i < batchCifNos.size(); i++) {
                    stringbuilder.append("?");
                    if (i < batchCifNos.size() - 1) {
                        stringbuilder.append(",");
                    }
                }
                stringbuilder.append(")");

                String checkPanData = stringbuilder.toString();
                PreparedStatement preparedstatement = conn.prepareStatement(checkPanData);

                // Set the CIF numbers as parameters
                for (int i = 0; i < batchCifNos.size(); i++) {
                    preparedstatement.setString(i + 1, batchCifNos.get(i));
                }

                ResultSet resultSet = preparedstatement.executeQuery();

                // Create a map to store CIF and their corresponding selected ID type
                // Store the PAN No in the map
                while (resultSet.next()) {
                    String cifNo = resultSet.getString("CIF_NO");
                    String id_No = resultSet.getString("ID_NO");
                    //  System.out.println("CIF No: " + cifNo);
                    String ID_Type = resultSet.getString("ID_TYPE");
                    //panavailable = true;
                    //   System.out.println("ID Type: " + ID_Type);
                    //  Map<String, ID_Type> panData = new HashMap<>();

//                    PanData.put(cifNo, ID_Type);
                    // Check if the CIF number is already processed
                    if (!PanData.containsKey(cifNo)) {
                        PanData.put(cifNo, new ID_Type(id_No, ID_Type));
                    } else {
                        // If the CIF number is already processed, prefer PAN
                        if ("PAN CARD".equalsIgnoreCase(ID_Type)) {
                            PanData.put(cifNo, new ID_Type(id_No, ID_Type)); // Overwrite if PAN is found
                        }
                    }
                }

                // print the map data
                for (Map.Entry<String, ID_Type> entry : PanData.entrySet()) {
                    System.out.println("CIF No: " + entry.getKey() + ", ID No: " + entry.getValue().getIdNo() + ", ID Type: " + entry.getValue().getIDType());
                }

            }
        } catch (Exception e) {
            e.printStackTrace();
        }
        return PanData;
    }

    //This is the method of GL Transaction
    public static void insrtGLTransaction(String KEY, String PACS_ID, double INTT_PAID, double tds_amount) {

        // String KEY_1 = null;
        double TXN_AMT = INTT_PAID - tds_amount;

        System.out.println("TXN Amount is : " + TXN_AMT);

        Double end_bal = 0.0;
        Double cum_curr_val = 0.0;
        Double tot_cr = 0.0;
        Double transaction_amount = 0.0;
        boolean rowUpdated = false;

        try (Connection conn = DriverManager.getConnection(URL, USERNAME, PASWORD)) {

            System.out.println("into the GL calculations method");
            String getBGLaccNo = "select GA.KEY_1 AS KEY_1 from GL_ACCOUNT GA,GL_PRODUCT GP where GP.ID = GA.GL_PROD_ID and GP.GL_CODE='37506' and GA.STATUS='A' and GA.PACS_ID = ? group by GA.KEY_1";

            PreparedStatement BGLpreparestatement = conn.prepareStatement(getBGLaccNo);

            BGLpreparestatement.setString(1, PACS_ID);

            ResultSet BGLresultSet = BGLpreparestatement.executeQuery();

            while (BGLresultSet.next()) {

                String KEY_1 = BGLresultSet.getString("KEY_1");

                System.out.println("Key_1 is : " + KEY_1);

//                if (KEY_1 == null) {
//                    System.out.println("No BGL avaiable for TDS of this PACS ID. Kindly check manually in the system.");
//                } 
                if (KEY_1.isEmpty()) {
                    System.out.println("No BGL avaiable for TDS of this PACS ID. Kindly check manually in the system.");
                } else {

                    long randomNumberGeneration = randomNumberGenerator();

                    //Fetch the End Balance and cumm_val from gl_txn and gl_account
                    String fetchLastRowSql = "Select COALESCE(END_BAL,TXN_AMT) from GL_TXN  where ACCT_NO =? ORDER BY POST_TIME DESC FETCH first 1 ROW ONLY";

                    PreparedStatement fetchpreparestatement = conn.prepareStatement(fetchLastRowSql);

                    fetchpreparestatement.setString(1, KEY_1);

                    ResultSet fetchrs = fetchpreparestatement.executeQuery();

                    if (fetchrs.next()) {
                        end_bal = fetchrs.getDouble("END_BAL");
                         transaction_amount = fetchrs.getDouble("TXN_AMT");
                        
                        System.out.println("end_bal is : " +end_bal);

                           if (end_bal == 0) {
                            end_bal = tds_amount+transaction_amount;
                        } else  {
                            end_bal = end_bal + tds_amount;
                        }
                    }else{
                        end_bal = transaction_amount;
                        System.out.println("No rows found. Setting end_bal to transaction_amount" + transaction_amount);
                    }

                    String fetchCumVal = "Select cum_curr_val,tot_cr from gl_account where key_1 = ? and pacs_id = ? and gl_prod_id ='061340427'";
                    PreparedStatement fetchCumValps = conn.prepareStatement(fetchCumVal);

                    fetchCumValps.setString(1, KEY_1);
                    fetchCumValps.setString(2, PACS_ID);

                    ResultSet fetchCumRs = fetchCumValps.executeQuery();

                    while (fetchCumRs.next()) {
                        cum_curr_val = fetchCumRs.getDouble("cum_curr_val");
                        tot_cr = fetchCumRs.getDouble("tot_cr");

                        if (cum_curr_val == 0) {
                            cum_curr_val = tds_amount;
                        } else {
                            cum_curr_val = cum_curr_val + tds_amount;
                        }

                        if (tot_cr == null) {
                            tot_cr = tds_amount;
                        } else {
                            tot_cr = tot_cr + tds_amount;
                        }
                    }

                    //Insert Into GL_TXN table
                    String insertGLTxn = "INSERT INTO GL_TXN (INST_NO,ACCT_NO,CBS_REF_NO,TRAN_DATE,POST_DATE,POST_TIME,JRNL_NO,TRAN_IND,END_BAL,TXN_AMT,NARRATION)\n"
                            + "values (?,?,?,?,?,?,?,?,?,?,?)";

                    PreparedStatement ps = conn.prepareStatement(insertGLTxn);
                    ps.setString(1, "003");
                    ps.setString(2, KEY_1);
                    ps.setString(3, KEY);
                    ps.setDate(4, java.sql.Date.valueOf(LocalDate.now()));
                    ps.setDate(5, java.sql.Date.valueOf(LocalDate.now()));
                    ps.setTimestamp(6, Timestamp.valueOf(LocalDateTime.now()));
                    ps.setLong(7, randomNumberGeneration);//randomNumberGeneration
                    ps.setString(8, "CR");
                    ps.setDouble(9, end_bal);
                    ps.setDouble(10, tds_amount);
                    ps.setString(11, "TDS Deducted and Credited");

                    ps.executeUpdate();

                    System.out.println("Insertion to GL TXN is completed having :" + KEY_1);

                    //Need to update the GL_ACCOUNT Table as well.
                    String updateGLAcc = "Update gl_account set cum_curr_val =? ,TOT_CR =? where KEY_1 =? and PACS_ID =? and GL_PROD_ID ='061340427'";

                    PreparedStatement updatepreparestatement = conn.prepareStatement(updateGLAcc);

                    updatepreparestatement.setDouble(1, end_bal);
                    updatepreparestatement.setDouble(2, tot_cr);
                    updatepreparestatement.setString(3, KEY_1);
                    updatepreparestatement.setString(4, PACS_ID);

                    rowUpdated = updatepreparestatement.executeUpdate() > 0;

                    System.out.println("Row: " + rowUpdated);
                }
            }

        } catch (Exception Ex) {
            Ex.printStackTrace();
        }

    }

    private static long randomNumberGenerator() {

        int currentYr = LocalDate.now().getYear();
        String yrPrfx = String.valueOf(currentYr);

        Random random = new Random();
        long digits = (long) (random.nextDouble() * 1_000_000_000_00L); // Generates up to 11 digits

        // Combine year prefix with random 11 digits
        String randomNumber = yrPrfx + String.format("%011d", digits);

        return Long.parseLong(randomNumber);

    }
}
